# Functions that return the correct URL that is going to be used to
# get annotation data from the desirable public data source.
#
# Copyright 2002, J. Zhang, all rights reserved.
#

getSrcUrl <- function(src = "LL", organism = "Homo sapiens", xml = TRUE,
                      dateOnly = FALSE){
    load(file.path(.path.package("AnnBuilder"), "data",
                    "sourceURLs.rda"), .GlobalEnv)
    switch(toupper(src),
           "LL" = return(getLLUrl()),
           "GP" = return(getUCSCUrl(organism)),
           "UG" = return(getUGUrl(organism)),
           "GO" = return(getGOUrl(xml, dateOnly)),
           "KEGG" = return(getKEGGUrl()),
           "GEO" = return(getGEOUrl()),
           "YG" = return(getYGUrl()),
           "HG" = return(getHGUrl()),
           "REFSEQ" = return(getRefSeqUrl(organism)),
           "ALL" = return(getAllUrl(organism)),
           stop(paste("Source", src, "is not supported.",
                      "Has to be LL, GP, UG, GO, KEGG, ALL")))
}

getAllUrl <- function(organism){
    urls <- c(LL = getLLUrl(), GP = getUCSCUrl(organism),
              UG = getUGUrl(organism), GO = getGOUrl(), KEGG = getKEGGUrl(),
              YG = getYGUrl(), HG = getHGUrl())
    return(urls)
}

getYGUrl <- function(){
    return("http://www.yeastgenome.org/DownloadContents.shtml")
}

getLLUrl <- function(){
    return(sourceURLs[["LL"]])
}

getUCSCUrl <- function(organism, downloadSite =
                       "http://hgdownload.cse.ucsc.edu/"){
    switch(toupper(organism),
           "HOMO SAPIENS" = key <- "goldenPath/hg.*Annotation database",
           "MUS MUSCULUS" = key <- "goldenPath/mm.*Annotation database",
           "RATTUS NORVEGICUS" = key <- "goldenPath/rn.*Annotation database",
           "DANIO RERIO" = key <- "goldenPath/danRer.*Annotation database",
           "CAENORHABDITIS ELEGANS" = key <-
                                      "goldenPath/ce.*Annotation database",
           "DROSOPHILA MELANOGASTER" = key <-
                                      "goldenPath/dm.*Annotation database",
           key <- NA)
    if(is.na(key)){
        warning(paste("Organism", organism, "is not supported by GP."))
        return(NA)
    }

    con <- url(sourceURLs[["GP"]], open = "r")
    htmlPage <- readLines(con)
    close(con)
    #lines <- htmlPage[grep(key, htmlPage)]
    got <- gsub(".*(goldenPath/.*/database/).*", "\\1",
                htmlPage[grep(key, htmlPage)])
    for(i in got){
        con <- url(paste(downloadSite, i, sep = ""))
        options(show.error.messages = FALSE, warn = -1)
        tryMe <- try(open(con))
        options(show.error.messages = TRUE, warn = 1)
        if(!inherits(tryMe, "try-error")){
            close(con)
            return(paste(downloadSite, i, sep = ""))
        }
        close(con)
    }
    warning("I could not get a valid url for Golden path")
    return(NA)
    #return(paste(gsub("(.*/).*.html", "\\1", sourceURLs[["GP"]]),
    #             got, sep = ""))
}

getUGUrl <- function(organism){
    #switch(toupper(organism),
    #       "HUMAN" = return(paste(sourceURLs[["UG"]], "/Hs.data.gz",
    #                        sep = "")),
    #       "MOUSE" = return(paste(sourceURLs[["UG"]], "/Mm.data.gz",
    #                        sep = "")),
    #       "RAT" = return(paste(sourceURLs[["UG"]],
    #                       "/Rn.data.gz", sep = "")),
    #       "ZEBRAFISH" = return(paste(sourceURLs[["UG"]], "/Dr.data.gz",
    #                      sep = "")),
    #       return(NA))
           #stop(paste("Organism", organism, "is not supported.")))

    return(paste(sourceURLs[["UG"]], "/", getUGShortName(organism),
                 ".data.gz", sep = ""))
}

# Figures out the url for the latest version of the GO data file
getGOUrl <- function(xml = TRUE, dateOnly = FALSE){
    # Read the html page to get the available data for different builds
    con <- url(sourceURLs[["GO"]], open = "r")
    srcHtml <-  readLines(con)
    close(con)
    builds <- srcHtml[grep("^<IMG SRC=.*>[0-9]*-[0-9]*-[0-9]*/.*", srcHtml)]
    # Loop through the builds to get the latest one
    builds <- sort(gsub(".*>([0-9]*-[0-9]*-[0-9]*)/.*", "\\1",  builds),
                   decreasing = TRUE)
    index <- 1
    # Get the name of the available file in the latest build
    repeat{
        con <- url(paste(sourceURLs[["GO"]], "/", builds[index],
                         sep = ""), open = "r")
        files <- readLines(con)
        close(con)
        if(xml){
            fileName <- gsub(".*(go_[0-9]*-termdb.xml.gz).*",
                    "\\1", files[grep(".*termdb.xml.gz", files)])
        }else{
            fileName <- gsub(".*(go_[0-9]*-termdb-data.gz).*", "\\1",
                             files[grep(".*termdb-dat", files)])
        }
        if(length(fileName) == 0){
            if(index < length(builds)){
                index <- index + 1
            }else{
                stop(paste("The target data file termdb.xxx.gz is not",
                           "available at", paste(sourceURLs[["GO"]],
                                                 "/", latest, sep = "")))
            }
        }else{
            if(dateOnly){
                return(builds[index])
            }else{
                return(paste(sourceURLs[["GO"]], "/",
                             builds[index], "/",fileName, sep = ""))
            }
        }
    }
}

getRefSeqUrl <- function(organism){
    switch(tolower(organism),
           "homo sapiens" = return("ftp://ftp.ncbi.nih.gov/genomes/H_sapiens/"),
           "yeast" =
    return("ftp://ftp.ncbi.nih.gov/genomes/Saccharomyces_cerevisiae/"),
           stop(paste("Organism", organism, "not supported")))
}

getKEGGUrl <- function(){
    return(sourceURLs[["KEGG"]])
}

getGEOUrl <- function(){
    return(sourceURLs[["GEO"]])
}

getHGUrl <- function(){
    return(sourceURLs[["HG"]])
}

getUGShortName <- function(sciName){
    temp <- UGSciNames()
    return(names(temp[tolower(temp) == tolower(sciName)]))
}


UGSciNames <- function(){
    c(Aga = "Anopheles gambiae", Ame = "Apis mellifera",
      Bmo = "Bombyx mori", Bt = "Bos taurus",
      Cel = "Caenorhabditis elegans", Cfa = "Canis familiaris",
      Cin = "Ciona intestinalis", Gga = "Gallus allus",
      Dm = "Drosophila melanogaster", Dr = "Danio rerio",
      Hma = "Hydra magnipapillata",
      Oar = "Ovis aries", Ola = "Oryzias latipes",
      Omy = "Oncorhynchus mykiss",
      Sma = "Schistosoma mansoni", Ssa = "Salmo salar",
      Ssc = "Sus scrofa", Str = "Xenopus tropicalis",
      Xl = "Xenopus laevis", At = "Arabidopsis thaliana",
      Gma = "Glycine max", Han = "Helianthus annus",
      Hv = "Hordeum vulgare",  Lsa = " Lactuca sativa",
      Les = "Lycopersicon esculentum", Lco = "Lotus corniculatus",
      Mdo = "Malus domestica", Mtr = "Medicago truncatula",
      Os = "Oryza sativa", Pta = "Pinus taeda", Ptp = "Populus tremula",
      Ptp = "Populus tremuloides", Sbi = "Sorghum bicolor",
      Sof = "Saccharum officinarum", Ta = "Triticum aestivum",
      Vva = "Vitis vinifera", Zm = "Zea mays",
      Cre = "Chlamydomonas reinhardtii", Ddi = "Dictyostelium discoideum",
      Mgr = "Magnaporthe grisea", Ncr = "Neurospora crassa",
      Tgo = "Toxoplasma gondii", Hs = "Homo sapiens", Mm ="Mus musculus",
      Rn = "Rattus norvegicus")
}

getShortSciName <- function(sciName){
    temp <- unlist(strsplit(sciName, " +"))
    if(length(temp) < 2){
        warning(paste("Species name may not be a two-word scientic",
                      "name.", sciName, "was returned."))
    }
    return(paste(toupper(substr(temp[1], 1, 1)),
                 substr(temp[2], 1, 2), sep = ""))
}
